package com.fwk.processor;

import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.util.Locale;

import javax.servlet.RequestDispatcher;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;

import com.fwk.actions.Action;
import com.fwk.actions.ActionForward;
import com.fwk.actions.ActionMessage;
import com.fwk.config.Configuration;
import com.fwk.config.actions.ActionConfig;
import com.fwk.config.exceptions.ExceptionConfig;
import com.fwk.config.forms.FormConfig;
import com.fwk.constants.GlobalConstants;
import com.fwk.controller.ServletController;
import com.fwk.exceptions.DefaultExceptionHandler;
import com.fwk.exceptions.ExceptionHandler;
import com.fwk.forms.FormBean;
import com.fwk.forms.FormErrors;

/**
 * S'encarrega de fer tot el tractament de la petici, validaci del formulari,
 * execuci de l'action associada i redirecci a la segent pantalla.
 * 
 * @author fjbravo
 * 
 */
public class Processor {

	// Referencia al servlet que ha delegat el tractament de la petici
	private ServletController servlet;

	// Configuraci del framework que ha de ser accessible per totes les
	// peticions
	private Configuration configuration;

	/**
	 * Mtode principal que s'encarrega de fer tot el tractament de la request
	 */
	public void process(HttpServletRequest req, HttpServletResponse res)
			throws IOException, ServletException {

		// 1) Busquem el path de l'acci
		String actionPath = processRequestPath(req, res);

		// 3) Modifiquem el Locale en cas que vingui informat a la request
		processLocale(req, res);

		// 2) Recuperem la configuraci de l'action associada a la petici
		ActionConfig actionConfig = processFindActionConfig(req, res,
				actionPath);
		if (actionConfig == null)
			return;
		// 3) Recuperem el formulari de l'acci. En el cas de tenir-lo associat
		FormBean formulari = processFindFormBean(req, res, actionConfig);
		if (formulari != null) {
			processFormPopulate(req, res, formulari, actionConfig);
			if (!processFormValidation(req, res, formulari, actionConfig)) {
				RequestDispatcher dispatcher = getServlet().getServletContext()
						.getRequestDispatcher("/");
				dispatcher.forward(req, res);
				return;
			}
		}

		Action action = processCreateAction(req, res, actionConfig);
		if (action == null)
			return;

		ActionForward forward = processExecuteAction(req, res, action,
				formulari, actionConfig);

		processExecuteForwarding(req, res, forward);
	}

	/**
	 * Mtode que canvia el locale per defecte de l'aplicaci segons el
	 * parmetre locale rebut per la request
	 * 
	 * @param req
	 * @param res
	 */
	protected void processLocale(HttpServletRequest req, HttpServletResponse res) {

		HttpSession session = req.getSession();

		if(req.getParameter("set-locale") != null){
			String localeName = req.getParameter("set-locale");
			String[] aux = localeName.split("_");
			Locale newLocale = new Locale(aux[0], aux[1]);
			session.setAttribute(GlobalConstants.GLOBAL_LOCALE, newLocale);
		}
		
		if(session.getAttribute(GlobalConstants.GLOBAL_LOCALE) == null){
			if (req.getLocale() != null) {
				session.setAttribute(GlobalConstants.GLOBAL_LOCALE, req.getLocale());
			}
			else
				session.setAttribute(GlobalConstants.GLOBAL_LOCALE, Locale.getDefault());
		}
	}

	/**
	 * Executa la redirecci a un recurs configurat a la action.
	 * 
	 * @param req
	 * @param res
	 * @param forward
	 */
	protected void processExecuteForwarding(HttpServletRequest req,
			HttpServletResponse res, ActionForward forward)
			throws ServletException, IOException {

		if (forward == null){
			forward = new ActionForward();
			forward.setAdreca("/");
		}
		

		String forwardPath = forward.getAdreca();

		if (forwardPath != null) {
			if(forwardPath.endsWith(".jsp") || forwardPath.endsWith(".html") || forwardPath.equals("/") ){
				RequestDispatcher dispatcher = getServlet().getServletContext()
					.getRequestDispatcher(forwardPath);
				dispatcher.forward(req, res);
			}
			else
				res.sendRedirect("/fwkTest" + forwardPath);
		} else
			return;
	}

	/**
	 * Executa la action associada a la petici.
	 * 
	 * @param req
	 * @param res
	 * @param action
	 * @param formulari
	 * @param actionConfig
	 */
	@SuppressWarnings("all")
	protected ActionForward processExecuteAction(HttpServletRequest req,
			HttpServletResponse res, Action action, FormBean formulari,
			ActionConfig actionConfig) throws ServletException, IOException {

		ActionForward forward = null;
		try {
			forward = action.execute(actionConfig, formulari, req, res);
			return forward;
		}
		// Aqu s on es realitzar el tractament de les Exceptions que llencen
		// les Action
		catch (Exception e) {
			forward = processException(e, req, res, action, formulari, actionConfig);
		}
		return forward;

	}

	/**
	 * Mtode que realitza el tractament d'una acci que es produeix en una
	 * acci
	 * 
	 * @param e
	 * @param req
	 * @param res
	 * @param action
	 * @param formulari
	 * @param actionConfig
	 * @return
	 * @throws ServletException
	 * @throws IOException
	 */
	@SuppressWarnings("all")
	protected ActionForward processException(Exception e, HttpServletRequest req,
			HttpServletResponse res, Action action, FormBean formulari,
			ActionConfig actionConfig) throws ServletException, IOException {

		ActionForward forward = null;
		String exceptionType = null;

		//En el cas que sigui una invocationTargetException mirarem la causa que la va originar
		if(e.getClass().getName().equals(InvocationTargetException.class.getName())){
			InvocationTargetException ivt = (InvocationTargetException) e;
			exceptionType = ivt.getTargetException().getClass().getCanonicalName();
		}
		else
			exceptionType = e.getClass().getCanonicalName();

		if (exceptionType != null) {
			// Primer busquem si tenim definida alguna acci en cas de
			// produir-se
			// una determinada excepci
			ExceptionConfig exConfig = actionConfig
					.findException(exceptionType);
			// Si tenim configuraci al respecte fem el tractament
			if (exConfig != null) {
				if (exConfig.getPath() != null
						&& !exConfig.getPath().equals("")) {
					forward = new ActionForward();
					forward.setAdreca(exConfig.getPath());
					ActionMessage msg = new ActionMessage(exConfig.getKey());
					FormErrors fe = new FormErrors();
					fe.addError("exception", msg);
					action.saveErrors(req, fe);
				}
				// Si no existeix la configuraci de l'acci buscarem un handler
				// associat
				// 1) Si el trobem, aquest ser l'encarregat de tractar l'acci
				// 2) Si no el trobem el tractar el handler per defecte
				else {
					ExceptionHandler handlerInstance = null;
					try {
						Class handlerClass = Class.forName(exConfig
								.getHandler());
						if (ExceptionHandler.class
								.isAssignableFrom(handlerClass)) {
							handlerInstance = (ExceptionHandler) handlerClass
									.newInstance();
						}
					} catch (Exception ex) {
						handlerInstance = new DefaultExceptionHandler();
					}
					if (handlerInstance != null) {
						return handlerInstance.execute(e, exConfig,
								actionConfig, formulari, req, res);
					}
				}
			}
			else{
				throw new ServletException(e.getStackTrace().toString());
			}
		}
		return forward;
	}

	/**
	 * Crea una nova acci configurada al fitxer de configuraci
	 * 
	 * @param req
	 * @param res
	 * @param actionConfig
	 * @return
	 */
	@SuppressWarnings("all")
	protected Action processCreateAction(HttpServletRequest req,
			HttpServletResponse res, ActionConfig actionConfig)
			throws IOException {

		Action actionInstance = null;
		String className = actionConfig.getType();

		try {
			Class actionClass = Class.forName(className);
			if (Action.class.isAssignableFrom(actionClass)) {
				actionInstance = (Action) actionClass.newInstance();
				return actionInstance;
			}
		} catch (Exception e) {
			res.sendError(HttpServletResponse.SC_INTERNAL_SERVER_ERROR, servlet
					.getInternalMessagesHandler().getProperty(
							"action.notFound", Locale.getDefault()));
		}
		return actionInstance;
	}

	/**
	 * Busca en l'objecte request el path de l'action que ha de tractar-la
	 * 
	 * @return
	 */
	protected String processRequestPath(HttpServletRequest req,
			HttpServletResponse res) {
		String actionPathName = req.getServletPath().replace(".do", "");
		return actionPathName;

	}

	/**
	 * Cerca la configuraci de l'acci identificada per el parmetre Path
	 * 
	 * @param req
	 * @param res
	 * @param actionPath
	 * @return
	 */
	protected ActionConfig processFindActionConfig(HttpServletRequest req,
			HttpServletResponse res, String actionPath) {
		return getConfiguration().getActionsMap().get(actionPath);
	}

	/**
	 * Busca el formulari associat a l'acci
	 * 
	 * @param req
	 * @param res
	 * @param actionConfig
	 * @return
	 */
	@SuppressWarnings("all")
	protected FormBean processFindFormBean(HttpServletRequest req,
			HttpServletResponse res, ActionConfig actionConfig)
			throws IOException, ServletException {
		FormBean formulari = null;
		String formName = actionConfig.getFormulari();
		if (formName == null)
			return null;
		else {
			FormConfig formConfig = getConfiguration().getFormsMap().get(
					formName);
			if (formConfig == null)
				return null;
			else {
				try {
					Class formClass = Class.forName(formConfig.getTipus());
					if (FormBean.class.isAssignableFrom(formClass)) {
						formulari = (FormBean) formClass.newInstance();
						req.setAttribute(formConfig.getNom(), formulari);
						return formulari;
					} else
						return null;
				} catch (Exception e) {
					return null;
				}

			}
		}
	}

	/**
	 * Executa el mtode validate del FormBean i retona true en cas de ser
	 * correcte
	 * 
	 * @param rq
	 * @param res
	 * @param form
	 * @return
	 */
	protected boolean processFormValidation(HttpServletRequest req,
			HttpServletResponse res, FormBean form, ActionConfig actionConfig) {
		boolean isValid = true;

		if (actionConfig.getValidate() != null
				&& actionConfig.getValidate().equals("true")) {
			FormErrors formErrors = form.validate();
			if (!formErrors.isEmpty())
				isValid = false;
			req.setAttribute("actionFormErrors", formErrors);
		}
		return isValid;
	}

	/**
	 * Mtode que realitza l'assignaci de valors provinents del formulari a
	 * l'objecte formBean
	 * 
	 * @param req
	 * @param res
	 * @param formulari
	 * @param actionConfig
	 */
	protected void processFormPopulate(HttpServletRequest req,
			HttpServletResponse res, FormBean formulari,
			ActionConfig actionConfig) {

		Field[] formFields = formulari.getClass().getDeclaredFields();
		for (Field field : formFields) {
			String reqParam = req.getParameter(field.getName());
			if (reqParam != null) {
				try {
					if (field.getType().equals(Integer.class)) {
						if (!reqParam.equals(""))
							field.set(formulari, Integer.parseInt(reqParam));
					} else if (field.getType().equals(Double.class)) {
						field
								.setDouble(formulari, Double
										.parseDouble(reqParam));
					} else if (field.getType().equals(Boolean.class)) {
						field.setBoolean(formulari, Boolean
								.parseBoolean(reqParam));
					} else
						field.set(formulari, reqParam);
				} catch (IllegalArgumentException e) {
					e.printStackTrace();
				} catch (IllegalAccessException e) {
					e.printStackTrace();
				}
			}
		}

	}

	/**
	 * Mtode per a inicialitzar la configuraci de la classe Processor
	 * 
	 * @param fwkServlet
	 * @param config
	 */
	public void initConfig(ServletController fwkServlet, Configuration config) {
		this.configuration = config;
		this.servlet = fwkServlet;

	}

	public ServletController getServlet() {
		return servlet;
	}

	public void setServlet(ServletController servlet) {
		this.servlet = servlet;
	}

	public Configuration getConfiguration() {
		return configuration;
	}

	public void setConfiguration(Configuration configuration) {
		this.configuration = configuration;
	}

}
